@@ -22,9 +22,9 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include "fcint.h"
 #include <dirent.h>
 #include <sys/types.h>
-#include "fcint.h"
 
 #if defined (_WIN32) && (defined (PIC) || defined (DLL_EXPORT))
 #define STRICT
@@ -77,40 +77,15 @@ FcConfigCreate (void)
     if (!config->rejectPatterns)
 	goto bail7;
 
-    config->cache = 0;
-    if (FcConfigHome())
-	if (!FcConfigSetCache (config, (FcChar8 *) ("~/" FC_USER_CACHE_FILE)))
-	    goto bail8;
-
-#ifdef _WIN32
-    if (config->cache == 0)
-    {
-	/* If no home, use the temp folder. */
-	FcChar8	    dummy[1];
-	int	    templen = GetTempPath (1, dummy);
-	FcChar8     *temp = malloc (templen + 1);
-
-	if (temp)
-	{
-	    FcChar8 *cache_dir;
-
-	    GetTempPath (templen + 1, temp);
-	    cache_dir = FcStrPlus (temp, FC_USER_CACHE_FILE);
-	    free (temp);
-	    if (!FcConfigSetCache (config, cache_dir))
-	    {
-		FcStrFree (cache_dir);
-		goto bail6;
-	    }
-	    FcStrFree (cache_dir);
-	}
-    }
-#endif
-
+    config->cacheDirs = FcStrSetCreate ();
+    if (!config->cacheDirs)
+	goto bail8;
+    
     config->blanks = 0;
 
     config->substPattern = 0;
     config->substFont = 0;
+    config->substScan = 0;
     config->maxObjects = 0;
     for (set = FcSetSystem; set <= FcSetApplication; set++)
 	config->fonts[set] = 0;
@@ -163,23 +138,10 @@ FcConfigNewestFile (FcStrSet *files)
     return newest;
 }
 
-FcFileTime
-FcConfigModifiedTime (FcConfig *config)
-{
-    if (!config)
-    {
-	FcFileTime v = { 0, FcFalse };
-	config = FcConfigGetCurrent ();
-	if (!config)
-	    return v;
-    }
-    return FcConfigNewestFile (config->configFiles);
-}
-
 FcBool
 FcConfigUptoDate (FcConfig *config)
 {
-    FcFileTime	config_time, font_time;
+    FcFileTime	config_time, config_dir_time, font_time;
     time_t	now = time(0);
     if (!config)
     {
@@ -188,8 +150,10 @@ FcConfigUptoDate (FcConfig *config)
 	    return FcFalse;
     }
     config_time = FcConfigNewestFile (config->configFiles);
+    config_dir_time = FcConfigNewestFile (config->configDirs);
     font_time = FcConfigNewestFile (config->fontDirs);
     if ((config_time.set && config_time.time - config->rescanTime > 0) ||
+	(config_dir_time.set && (config_dir_time.time - config->rescanTime) > 0) ||
 	(font_time.set && (font_time.time - config->rescanTime) > 0))
     {
 	return FcFalse;
@@ -226,6 +190,7 @@ FcConfigDestroy (FcConfig *config)
 
     FcStrSetDestroy (config->configDirs);
     FcStrSetDestroy (config->fontDirs);
+    FcStrSetDestroy (config->cacheDirs);
     FcStrSetDestroy (config->configFiles);
     FcStrSetDestroy (config->acceptGlobs);
     FcStrSetDestroy (config->rejectGlobs);
@@ -235,11 +200,9 @@ FcConfigDestroy (FcConfig *config)
     if (config->blanks)
 	FcBlanksDestroy (config->blanks);
 
-    if (config->cache)
-	FcStrFree (config->cache);
-
     FcSubstDestroy (config->substPattern);
     FcSubstDestroy (config->substFont);
+    FcSubstDestroy (config->substScan);
     for (set = FcSetSystem; set <= FcSetApplication; set++)
 	if (config->fonts[set])
 	    FcFontSetDestroy (config->fonts[set]);
@@ -248,103 +211,130 @@ FcConfigDestroy (FcConfig *config)
     FcMemFree (FC_MEM_CONFIG, sizeof (FcConfig));
 }
 
+/*
+ * Add cache to configuration, adding fonts and directories
+ */
+
+FcBool
+FcConfigAddCache (FcConfig *config, FcCache *cache, 
+		  FcSetName set, FcStrSet *dirSet)
+{
+    FcFontSet	*fs;
+    intptr_t	*dirs;
+    int		i;
+
+    /*
+     * Add fonts
+     */
+    fs = FcCacheSet (cache);
+    if (fs)
+    {
+	int	nref = 0;
+	
+	for (i = 0; i < fs->nfont; i++)
+	{
+	    FcPattern	*font = FcFontSetFont (fs, i);
+	    FcChar8	*font_file;
+
+	    /*
+	     * Check to see if font is banned by filename
+	     */
+	    if (FcPatternObjectGetString (font, FC_FILE_OBJECT,
+					  0, &font_file) == FcResultMatch &&
+		!FcConfigAcceptFilename (config, font_file))
+	    {
+		continue;
+	    }
+		
+	    /*
+	     * Check to see if font is banned by pattern
+	     */
+	    if (!FcConfigAcceptFont (config, font))
+		continue;
+		
+	    nref++;
+	    FcFontSetAdd (config->fonts[set], font);
+	}
+	FcDirCacheReference (cache, nref);
+    }
+
+    /*
+     * Add directories
+     */
+    dirs = FcCacheDirs (cache);
+    if (dirs)
+    {
+	for (i = 0; i < cache->dirs_count; i++)
+	{
+	    FcChar8	*dir = FcOffsetToPtr (dirs, dirs[i], FcChar8);
+	    if (FcConfigAcceptFilename (config, dir))
+		FcStrSetAddFilename (dirSet, dir);
+	}
+    }
+    return FcTrue;
+}
+
+static FcBool
+FcConfigAddDirList (FcConfig *config, FcSetName set, FcStrSet *dirSet)
+{
+    FcStrList	    *dirlist;
+    FcChar8	    *dir;
+    FcCache	    *cache;
+    
+    dirlist = FcStrListCreate (dirSet);
+    if (!dirlist)
+        return FcFalse;
+	
+    while ((dir = FcStrListNext (dirlist)))
+    {
+	if (FcDebug () & FC_DBG_FONTSET)
+	    printf ("adding fonts from%s\n", dir);
+	cache = FcDirCacheRead (dir, FcFalse, config);
+	if (!cache)
+	    continue;
+	FcConfigAddCache (config, cache, set, dirSet);
+	FcDirCacheUnload (cache);
+    }
+    FcStrListDone (dirlist);
+    return FcTrue;
+}
+
 /*
  * Scan the current list of directories in the configuration
- * and build the set of available fonts. Update the
- * per-user cache file to reflect the new configuration
+ * and build the set of available fonts.
  */
 
 FcBool
 FcConfigBuildFonts (FcConfig *config)
 {
-    FcFontSet	    *fonts, *cached_fonts;
-    FcGlobalCache   *cache;
-    FcStrList	    *list;
-    FcStrSet	    *oldDirs;
-    FcChar8	    *dir;
+    FcFontSet	    *fonts;
 
+    if (!config)
+    {
+	config = FcConfigGetCurrent ();
+	if (!config)
+	    return FcFalse;
+    }
+	
     fonts = FcFontSetCreate ();
     if (!fonts)
-	goto bail0;
+	return FcFalse;
     
-    cache = FcGlobalCacheCreate ();
-    if (!cache)
-	goto bail1;
-
-    oldDirs = FcStrSetCreate ();
-    if (!oldDirs)
-        goto bail2;
-
-    if (config->cache)
-	FcGlobalCacheLoad (cache, oldDirs, config->cache, config);
-
-    cached_fonts = FcCacheRead(config, cache);
-    if (!cached_fonts)
-    {
-	list = FcConfigGetFontDirs (config);
-	if (!list)
-	    goto bail2;
-	
-	while ((dir = FcStrListNext (list)))
-	{
-	    if (FcDebug () & FC_DBG_FONTSET)
-		printf ("build scan dir %s\n", dir);
-	    FcDirScanConfig (fonts, config->fontDirs, cache, 
-			     config->blanks, dir, FcFalse, config);
-	}
-	
-	FcStrListDone (list);
-    }
-    else
-    {
-	int i;
-
-        for (i = 0; i < oldDirs->num; i++)
-        {
-	    if (FcDebug () & FC_DBG_FONTSET)
-		printf ("scan dir %s\n", oldDirs->strs[i]);
-	    FcDirScanConfig (fonts, config->fontDirs, cache, 
-			     config->blanks, oldDirs->strs[i], 
-                             FcFalse, config);
-	}
-
-	for (i = 0; i < cached_fonts->nfont; i++)
-	{
-	    FcChar8 	*cfn; 
-	    FcPatternGetString (cached_fonts->fonts[i], FC_FILE, 0, &cfn);
-
-	    if (FcConfigAcceptFont (config, cached_fonts->fonts[i]) &&
-                (cfn && FcConfigAcceptFilename (config, cfn)))
-		FcFontSetAdd (fonts, cached_fonts->fonts[i]);
-
-	    cached_fonts->fonts[i] = 0; /* prevent free in FcFontSetDestroy */
-	}
-	cached_fonts->nfont = 0;
-	FcFontSetDestroy (cached_fonts);
-    }
-    
-    if (FcDebug () & FC_DBG_FONTSET)
-	FcFontSetPrint (fonts);
-
-    if (config->cache)
-	FcGlobalCacheSave (cache, config->cache, config);
-    FcGlobalCacheDestroy (cache);
-    FcStrSetDestroy (oldDirs);
-
     FcConfigSetFonts (config, fonts, FcSetSystem);
     
+    if (!FcConfigAddDirList (config, FcSetSystem, config->fontDirs))
+	return FcFalse;
+    if (FcDebug () & FC_DBG_FONTSET)
+	FcFontSetPrint (fonts);
     return FcTrue;
-bail2:
-    FcStrSetDestroy (oldDirs);
-bail1:
-    FcFontSetDestroy (fonts);
-bail0:
-    return FcFalse;
 }
 
 FcBool
 FcConfigSetCurrent (FcConfig *config)
 {
+    if (config == _fcConfig)
+	return FcTrue;
+
     if (!config->fonts)
 	if (!FcConfigBuildFonts (config))
 	    return FcFalse;
@@ -383,113 +373,13 @@ FcConfigGetConfigDirs (FcConfig   *config)
     return FcStrListCreate (config->configDirs);
 }
 
-static FcChar8 *
-FcConfigInodeMatchFontDir (FcConfig *config, const FcChar8 *d)
-{
-    int		n;
-    ino_t	di;
-    dev_t	dd;
-    struct stat s;
-
-    /* first we do string matches rather than file accesses */
-    /* FcStrSetMember doesn't tell us the index so that we can return
-     * the config-owned copy. */
-    for (n = 0; n < config->fontDirs->num; n++)
-    {
-	if (!FcStrCmp (config->fontDirs->strs[n], d))
-	    return config->fontDirs->strs[n];
-    }
-
-    /* If this is a bottleneck, we can cache the fontDir inodes. */
-    if (stat ((char *)d, &s) == -1)
-	return 0;
-    di = s.st_ino; dd = s.st_dev;
-
-    for (n = 0; n < config->fontDirs->num; n++)
-    {
-	if (stat ((char *)config->fontDirs->strs[n], &s) == -1)
-	    continue;
-	if (di == s.st_ino && dd == s.st_dev)
-	    return config->fontDirs->strs[n];
-    }
-    return 0;
-}
-
 FcBool
 FcConfigAddFontDir (FcConfig	    *config,
 		    const FcChar8   *d)
 {
-    /* Avoid adding d if it's an alias of something else, too. */
-    if (FcConfigInodeMatchFontDir(config, d))
-	return FcTrue;
     return FcStrSetAddFilename (config->fontDirs, d);
 }
 
-static FcBool
-FcConfigAddFontDirSubdirs (FcConfig        *config,
-			   const FcChar8   *d)
-{
-    DIR *dir;
-    struct dirent *e;
-    FcChar8 *subdir;
-    FcBool added = FcFalse;
-    
-    if (!(dir = opendir ((char *) d)))
-	return FcFalse;
-    if (!(subdir = (FcChar8 *) malloc (strlen ((char *) d) + FC_MAX_FILE_LEN + 2)))
-    {
-	fprintf (stderr, "out of memory");
-	return FcFalse;
-    }
-    while ((e = readdir (dir)))
-    {
-	if (strcmp (e->d_name, ".") && strcmp (e->d_name, "..") &&
-	    strlen (e->d_name) < FC_MAX_FILE_LEN)
-	{
-	    strcpy ((char *)subdir, (char *)d);
-	    strcat ((char *)subdir, "/");
-	    strcat ((char *)subdir, e->d_name);
-	    if (FcFileIsDir (subdir))
-	    {
-		if (FcConfigInodeMatchFontDir(config, subdir))
-		    continue; /* already added */
-		FcStrSetAddFilename (config->fontDirs, subdir);
-		FcConfigAddFontDirSubdirs (config, subdir);
-		added = FcTrue;
-	    }
-	}
-    }
-    free (subdir);
-    closedir (dir);
-    return added;
-}
-
-const FcChar8 *
-FcConfigNormalizeFontDir (FcConfig  	*config, 
-			  const FcChar8 *d)
-{
-    FcChar8 	*d0;
-    int		n, n0;
-    FcBool 	added = FcFalse;
-
-    d0 = FcConfigInodeMatchFontDir(config, d);
-    if (d0)
-	return d0;
-
-    /* Ok, we didn't find it in fontDirs; let's add subdirs.... */
-    for (n = 0, n0 = config->fontDirs->num; n < n0; n++) 
-    {
-	if (FcConfigAddFontDirSubdirs (config, config->fontDirs->strs[n]))
-	    added = FcTrue;
-    }
-
-    /* ... and try again. */
-    if (added)
-	return FcConfigInodeMatchFontDir(config, d);
-
-    return 0;
-}
-
 FcBool
 FcConfigAddDir (FcConfig	    *config,
 		const FcChar8	    *d)
@@ -510,6 +400,25 @@ FcConfigGetFontDirs (FcConfig	*config)
     return FcStrListCreate (config->fontDirs);
 }
 
+FcBool
+FcConfigAddCacheDir (FcConfig	    *config,
+		     const FcChar8  *d)
+{
+    return FcStrSetAddFilename (config->cacheDirs, d);
+}
+
+FcStrList *
+FcConfigGetCacheDirs (FcConfig	*config)
+{
+    if (!config)
+    {
+	config = FcConfigGetCurrent ();
+	if (!config)
+	    return 0;
+    }
+    return FcStrListCreate (config->cacheDirs);
+}
+    
 FcBool
 FcConfigAddConfigFile (FcConfig	    *config,
 		       const FcChar8   *f)
@@ -537,30 +446,10 @@ FcConfigGetConfigFiles (FcConfig    *config)
     return FcStrListCreate (config->configFiles);
 }
 
-FcBool
-FcConfigSetCache (FcConfig	*config,
-		  const FcChar8	*c)
-{
-    FcChar8    *new = FcStrCopyFilename (c);
-    
-    if (!new)
-	return FcFalse;
-    if (config->cache)
-	FcStrFree (config->cache);
-    config->cache = new;
-    return FcTrue;
-}
-
 FcChar8 *
 FcConfigGetCache (FcConfig  *config)
 {
-    if (!config)
-    {
-	config = FcConfigGetCurrent ();
-	if (!config)
-	    return 0;
-    }
-    return config->cache;
+    return NULL;
 }
 
 FcFontSet *
@@ -586,8 +475,6 @@ FcConfigSetFonts (FcConfig	*config,
     config->fonts[set] = fonts;
 }
 
-
-
 FcBlanks *
 FcConfigGetBlanks (FcConfig	*config)
 {
@@ -604,23 +491,27 @@ FcBool
 FcConfigAddBlank (FcConfig	*config,
 		  FcChar32    	blank)
 {
-    FcBlanks	*b;
+    FcBlanks	*b, *freeme = 0;
     
     b = config->blanks;
     if (!b)
     {
-	b = FcBlanksCreate ();
+	freeme = b = FcBlanksCreate ();
 	if (!b)
 	    return FcFalse;
     }
     if (!FcBlanksAdd (b, blank))
+    {
+        if (freeme)
+            FcBlanksDestroy (freeme);
 	return FcFalse;
+    }
     config->blanks = b;
     return FcTrue;
 }
 
 int
-FcConfigGetRescanInverval (FcConfig *config)
+FcConfigGetRescanInterval (FcConfig *config)
 {
     if (!config)
     {
@@ -632,7 +523,7 @@ FcConfigGetRescanInverval (FcConfig *config)
 }
 
 FcBool
-FcConfigSetRescanInverval (FcConfig *config, int rescanInterval)
+FcConfigSetRescanInterval (FcConfig *config, int rescanInterval)
 {
     if (!config)
     {
@@ -644,6 +535,22 @@ FcConfigSetRescanInverval (FcConfig *config, int rescanInterval)
     return FcTrue;
 }
 
+/*
+ * A couple of typos escaped into the library
+ */
+int
+FcConfigGetRescanInverval (FcConfig *config)
+{
+    return FcConfigGetRescanInterval (config);
+}
+
+FcBool
+FcConfigSetRescanInverval (FcConfig *config, int rescanInterval)
+{
+    return FcConfigSetRescanInterval (config, rescanInterval);
+}
+
+    
 FcBool
 FcConfigAddEdit (FcConfig	*config,
 		 FcTest		*test,
@@ -654,14 +561,23 @@ FcConfigAddEdit (FcConfig	*config,
     FcTest	*t;
     int		num;
 
+    switch (kind) {
+    case FcMatchPattern:
+	prev = &config->substPattern;
+	break;
+    case FcMatchFont:
+	prev = &config->substFont;
+	break;
+    case FcMatchScan:
+	prev = &config->substScan;
+	break;
+    default:
+	return FcFalse;
+    }
     subst = (FcSubst *) malloc (sizeof (FcSubst));
     if (!subst)
 	return FcFalse;
     FcMemAlloc (FC_MEM_SUBST, sizeof (FcSubst));
-    if (kind == FcMatchPattern)
-	prev = &config->substPattern;
-    else
-	prev = &config->substFont;
     for (; *prev; prev = &(*prev)->next);
     *prev = subst;
     subst->next = 0;
@@ -892,6 +808,7 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
     FcValue	v, vl, vr;
     FcResult	r;
     FcMatrix	*m;
+    FcChar8     *str;
     
     switch (e->op) {
     case FcOpInteger:
@@ -921,7 +838,7 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 	v.u.b = e->u.bval;
 	break;
     case FcOpField:
-	r = FcPatternGet (p, e->u.field, 0, &v);
+	r = FcPatternObjectGet (p, e->u.object, 0, &v);
 	if (r != FcResultMatch)
 	    v.type = FcTypeVoid;
 	v = FcValueSave (v);
@@ -1022,7 +939,9 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 		switch (e->op) {
 		case FcOpPlus:
 		    v.type = FcTypeString;
-		    v.u.s = FcStrStaticName (FcStrPlus (vl.u.s, vr.u.s));
+		    str = FcStrPlus (vl.u.s, vr.u.s);
+		    v.u.s = FcStrStaticName (str);
+		    FcStrFree (str);
 			 
 		    if (!v.u.s)
 			v.type = FcTypeVoid;
@@ -1171,7 +1090,7 @@ FcConfigMatchValueList (FcPattern	*p,
 	    e = 0;
 	}
 
-	for (v = values; v; v = FcValueListPtrU(v->next))
+	for (v = values; v; v = FcValueListNext(v))
 	{
 	    /* Compare the pattern value to the match expression value */
 	    if (FcConfigCompareValue (&v->value, t->op, &value))
@@ -1207,17 +1126,17 @@ FcConfigValues (FcPattern *p, FcExpr *e, FcValueBinding binding)
     if (e->op == FcOpComma)
     {
 	l->value = FcConfigEvaluate (p, e->u.tree.left);
-	l->next  = FcValueListPtrCreateDynamic(FcConfigValues (p, e->u.tree.right, binding));
+	l->next = FcConfigValues (p, e->u.tree.right, binding);
     }
     else
     {
 	l->value = FcConfigEvaluate (p, e);
-	l->next  = FcValueListPtrCreateDynamic(0);
+	l->next = NULL;
     }
     l->binding = binding;
     if (l->value.type == FcTypeVoid)
     {
-	FcValueList  *next = FcValueListPtrU(l->next);
+	FcValueList  *next = FcValueListNext(l);
 
 	FcMemFree (FC_MEM_VALLIST, sizeof (FcValueList));
 	free (l);
@@ -1240,27 +1159,26 @@ FcConfigAdd (FcValueListPtr *head,
 	sameBinding = position->binding;
     else
 	sameBinding = FcValueBindingWeak;
-    for (v = FcValueListPtrCreateDynamic(new); FcValueListPtrU(v); 
-	 v = FcValueListPtrU(v)->next)
-	if (FcValueListPtrU(v)->binding == FcValueBindingSame)
-	    FcValueListPtrU(v)->binding = sameBinding;
+    for (v = new; v != NULL; v = FcValueListNext(v))
+	if (v->binding == FcValueBindingSame)
+	    v->binding = sameBinding;
     if (append)
     {
 	if (position)
 	    prev = &position->next;
 	else
-	    for (prev = head; FcValueListPtrU(*prev); 
-		 prev = &(FcValueListPtrU(*prev)->next))
+	    for (prev = head; *prev != NULL; 
+		 prev = &(*prev)->next)
 		;
     }
     else
     {
 	if (position)
 	{
-	    for (prev = head; FcValueListPtrU(*prev); 
-		 prev = &(FcValueListPtrU(*prev)->next))
+	    for (prev = head; *prev != NULL; 
+		 prev = &(*prev)->next)
 	    {
-		if (FcValueListPtrU(*prev) == position)
+		if (*prev == position)
 		    break;
 	    }
 	}
@@ -1269,7 +1187,7 @@ FcConfigAdd (FcValueListPtr *head,
 
 	if (FcDebug () & FC_DBG_EDIT)
 	{
-	    if (!FcValueListPtrU(*prev))
+	    if (*prev == NULL)
 		printf ("position not on list\n");
 	}
     }
@@ -1283,12 +1201,12 @@ FcConfigAdd (FcValueListPtr *head,
     
     if (new)
     {
-	last = FcValueListPtrCreateDynamic(new);
-	while (FcValueListPtrU(FcValueListPtrU(last)->next))
-	    last = FcValueListPtrU(last)->next;
+	last = new;
+	while (last->next != NULL)
+	    last = last->next;
     
-	FcValueListPtrU(last)->next = *prev;
-	*prev = FcValueListPtrCreateDynamic(new);
+	last->next = *prev;
+	*prev = new;
     }
     
     if (FcDebug () & FC_DBG_EDIT)
@@ -1307,14 +1225,13 @@ FcConfigDel (FcValueListPtr *head,
 {
     FcValueListPtr *prev;
 
-    for (prev = head; FcValueListPtrU(*prev); 
-	 prev = &(FcValueListPtrU(*prev)->next))
+    for (prev = head; *prev != NULL; prev = &(*prev)->next)
     {
-	if (FcValueListPtrU(*prev) == position)
+	if (*prev == position)
 	{
 	    *prev = position->next;
-	    position->next = FcValueListPtrCreateDynamic(0);
-	    FcValueListDestroy (FcValueListPtrCreateDynamic(position));
+	    position->next = NULL;
+	    FcValueListDestroy (position);
 	    break;
 	}
     }
@@ -1322,13 +1239,13 @@ FcConfigDel (FcValueListPtr *head,
 
 static void
 FcConfigPatternAdd (FcPattern	*p,
-		    const char	*object,
+		    FcObject	object,
 		    FcValueList	*list,
 		    FcBool	append)
 {
     if (list)
     {
-	FcPatternElt    *e = FcPatternInsertElt (p, object);
+	FcPatternElt    *e = FcPatternObjectInsertElt (p, object);
     
 	if (!e)
 	    return;
@@ -1341,24 +1258,24 @@ FcConfigPatternAdd (FcPattern	*p,
  */
 static void
 FcConfigPatternDel (FcPattern	*p,
-		    const char	*object)
+		    FcObject	object)
 {
-    FcPatternElt    *e = FcPatternFindElt (p, object);
+    FcPatternElt    *e = FcPatternObjectFindElt (p, object);
     if (!e)
 	return;
-    while (FcValueListPtrU(e->values))
-	FcConfigDel (&e->values, FcValueListPtrU(e->values));
+    while (e->values != NULL)
+	FcConfigDel (&e->values, e->values);
 }
 
 static void
 FcConfigPatternCanon (FcPattern	    *p,
-		      const char    *object)
+		      FcObject	    object)
 {
-    FcPatternElt    *e = FcPatternFindElt (p, object);
+    FcPatternElt    *e = FcPatternObjectFindElt (p, object);
     if (!e)
 	return;
-    if (!FcValueListPtrU(e->values))
-	FcPatternDel (p, object);
+    if (e->values == NULL)
+	FcPatternObjectDel (p, object);
 }
 
 FcBool
@@ -1382,6 +1299,20 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	    return FcFalse;
     }
 
+    switch (kind) {
+    case FcMatchPattern:
+	s = config->substPattern;
+	break;
+    case FcMatchFont:
+	s = config->substFont;
+	break;
+    case FcMatchScan:
+	s = config->substScan;
+	break;
+    default:
+	return FcFalse;
+    }
+
     st = (FcSubState *) malloc (config->maxObjects * sizeof (FcSubState));
     if (!st && config->maxObjects)
 	return FcFalse;
@@ -1392,10 +1323,6 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	printf ("FcConfigSubstitute ");
 	FcPatternPrint (p);
     }
-    if (kind == FcMatchPattern)
-	s = config->substPattern;
-    else
-	s = config->substFont;
     for (; s; s = s->next)
     {
 	/*
@@ -1415,7 +1342,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	    else
 		m = p;
 	    if (m)
-		st[i].elt = FcPatternFindElt (m, t->field);
+		st[i].elt = FcPatternObjectFindElt (m, t->object);
 	    else
 		st[i].elt = 0;
 	    /*
@@ -1436,12 +1363,12 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	     * Check to see if there is a match, mark the location
 	     * to apply match-relative edits
 	     */
-	    st[i].value = FcConfigMatchValueList (m, t, FcValueListPtrU(st[i].elt->values));
+	    st[i].value = FcConfigMatchValueList (m, t, st[i].elt->values);
 	    if (!st[i].value)
 		break;
-	    if (t->qual == FcQualFirst && st[i].value != FcValueListPtrU(st[i].elt->values))
+	    if (t->qual == FcQualFirst && st[i].value != st[i].elt->values)
 		break;
-	    if (t->qual == FcQualNotFirst && st[i].value == FcValueListPtrU(st[i].elt->values))
+	    if (t->qual == FcQualNotFirst && st[i].value == st[i].elt->values)
 		break;
 	}
 	if (t)
@@ -1469,8 +1396,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	    for (t = s->test, i = 0; t; t = t->next, i++)
 	    {
 		if ((t->kind == FcMatchFont || kind == FcMatchPattern) &&
-		    !FcStrCmpIgnoreCase ((FcChar8 *) t->field, 
-					 (FcChar8 *) e->field))
+		    t->object == e->object)
 		{
 		    /* 
 		     * KLUDGE - the pattern may have been reallocated or
@@ -1479,7 +1405,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		     * the element again
 		     */
 		    if (e != s->edit && st[i].elt)
-			st[i].elt = FcPatternFindElt (p, t->field);
+			st[i].elt = FcPatternObjectFindElt (p, t->object);
 		    if (!st[i].elt)
 			t = 0;
 		    break;
@@ -1494,7 +1420,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		if (t)
 		{
 		    FcValueList	*thisValue = st[i].value;
-		    FcValueList	*nextValue = thisValue ? FcValueListPtrU(thisValue->next) : 0;
+		    FcValueList	*nextValue = thisValue;
 		    
 		    /*
 		     * Append the new list of values after the current value
@@ -1503,7 +1429,8 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		    /*
 		     * Delete the marked value
 		     */
-		    FcConfigDel (&st[i].elt->values, thisValue);
+                    if (thisValue)
+			FcConfigDel (&st[i].elt->values, thisValue);
 		    /*
 		     * Adjust any pointers into the value list to ensure
 		     * future edits occur at the same place
@@ -1521,8 +1448,8 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		 * Delete all of the values and insert
 		 * the new set
 		 */
-		FcConfigPatternDel (p, e->field);
-		FcConfigPatternAdd (p, e->field, l, FcTrue);
+		FcConfigPatternDel (p, e->object);
+		FcConfigPatternAdd (p, e->object, l, FcTrue);
 		/*
 		 * Adjust any pointers into the value list as they no
 		 * longer point to anything valid
@@ -1545,7 +1472,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		}
 		/* fall through ... */
 	    case FcOpPrependFirst:
-		FcConfigPatternAdd (p, e->field, l, FcFalse);
+		FcConfigPatternAdd (p, e->object, l, FcFalse);
 		break;
 	    case FcOpAppend:
 		if (t)
@@ -1555,9 +1482,10 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		}
 		/* fall through ... */
 	    case FcOpAppendLast:
-		FcConfigPatternAdd (p, e->field, l, FcTrue);
+		FcConfigPatternAdd (p, e->object, l, FcTrue);
 		break;
 	    default:
+                FcValueListDestroy (l);
 		break;
 	    }
 	}
@@ -1566,7 +1494,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	 * any properties without data
 	 */
 	for (e = s->edit; e; e = e->next)
-	    FcConfigPatternCanon (p, e->field);
+	    FcConfigPatternCanon (p, e->object);
 
 	if (FcDebug () & FC_DBG_EDIT)
 	{
@@ -1862,7 +1790,7 @@ FcConfigAppFontAddFile (FcConfig    *config,
 	FcConfigSetFonts (config, set, FcSetApplication);
     }
 	
-    if (!FcFileScanConfig (set, subdirs, 0, config->blanks, file, FcFalse, config))
+    if (!FcFileScanConfig (set, subdirs, config->blanks, file, config))
     {
 	FcStrSetDestroy (subdirs);
 	return FcFalse;
@@ -1884,9 +1812,7 @@ FcConfigAppFontAddDir (FcConfig	    *config,
 		       const FcChar8   *dir)
 {
     FcFontSet	*set;
-    FcStrSet	*subdirs;
-    FcStrList	*sublist;
-    FcChar8	*subdir;
+    FcStrSet	*dirs;
     
     if (!config)
     {
@@ -1894,8 +1820,9 @@ FcConfigAppFontAddDir (FcConfig	    *config,
 	if (!config)
 	    return FcFalse;
     }
-    subdirs = FcStrSetCreate ();
-    if (!subdirs)
+
+    dirs = FcStrSetCreate ();
+    if (!dirs)
 	return FcFalse;
     
     set = FcConfigGetFonts (config, FcSetApplication);
@@ -1904,26 +1831,20 @@ FcConfigAppFontAddDir (FcConfig	    *config,
 	set = FcFontSetCreate ();
 	if (!set)
 	{
-	    FcStrSetDestroy (subdirs);
+	    FcStrSetDestroy (dirs);
 	    return FcFalse;
 	}
 	FcConfigSetFonts (config, set, FcSetApplication);
     }
     
-    if (!FcDirScanConfig (set, subdirs, 0, config->blanks, dir, FcFalse, config))
+    FcStrSetAddFilename (dirs, dir);
+    
+    if (!FcConfigAddDirList (config, FcSetApplication, dirs))
     {
-	FcStrSetDestroy (subdirs);
+	FcStrSetDestroy (dirs);
 	return FcFalse;
     }
-    if ((sublist = FcStrListCreate (subdirs)))
-    {
-	while ((subdir = FcStrListNext (sublist)))
-	{
-	    FcConfigAppFontAddDir (config, subdir);
-	}
-	FcStrListDone (sublist);
-    }
-    FcStrSetDestroy (subdirs);
+    FcStrSetDestroy (dirs);
     return FcTrue;
 }
 
@@ -2049,3 +1970,6 @@ FcConfigAcceptFont (FcConfig	    *config,
 	return FcFalse;
     return FcTrue;
 }
+#define __fccfg__
+#include "fcaliastail.h"
+#undef __fccfg__
